﻿/*
 * How AddForce() works:
 * ---------------------
 * 
 * - Add force modifies the velocity during the physics update. The force mode
 *   determines *how* the velocity changes each time step.
 *   
 * - ForceMode.VelocityChange interprets the vector as a velocity directly.
 *   This is what the Unity documentation means when it says that "the unit of
 *   the force parameter is applied to the rigidbody as distance/time."
 *   AddForce(v, ForceMode.VelocityChange) will add v to the velocity for the
 *   current physics time step. If this is called each time step, the body's
 *   velocity will increase by v each time. Below, the magnitude of v is 0.15.
 *   Note that the velocity of the body increases by 0.15 each physics update
 *   and stays constant throughout the graphics updates.
 *
 *    Fixed Update: AddForce (0.13, 0.00, -0.07)
 *          Update: velocity magnitude=0.15
 *    Fixed Update: AddForce (0.13, 0.00, -0.07)
 *          Update: velocity magnitude=0.30
 *          Update: velocity magnitude=0.30
 *          Update: velocity magnitude=0.30
 *    Fixed Update: AddForce (0.13, 0.00, -0.07)
 *          Update: velocity magnitude=0.45
 *
 * - ForceMode.Acceleration interprets the vector as distance/time^2, which are
 *   the units of acceleration. To compute the velocity to be added in the
 *   current time step, acceleration must of course be multiplied by the time
 *   step, resulting in units of distance/time (a velocity, just as in 
 *   ForceMode.VelocityChange). Below, the time step is 1/50 = .02. The vector
 *   v is again 0.15, so each physics update that AddForce() is called, 
 *   0.15 * .02 = .003 is added to the velocity.
 *
 *    Fixed Update: AddForce (0.1482, 0.0000, -0.0231)
 *    Fixed Update: AddForce (0.1482, 0.0000, -0.0231)
 *    Fixed Update: AddForce (0.1482, 0.0000, -0.0231)
 *          Update: velocity magnitude=0.0090 <-- 3 * .003 = .009
 *          Update: velocity magnitude=0.0090
 *    Fixed Update: AddForce (0.1482, 0.0000, -0.0231)
 *          Update: velocity magnitude=0.0120
 *          Update: velocity magnitude=0.0120
 *          Update: velocity magnitude=0.0120
 *
 * - ForceMode.Impulse is the same as ForceMode.VelocityChange except that the
 *   units include mass: mass*distance/time. Therefore, the velocity change for
 *   the physics update is found simply by dividing by the rigid body's mass.
 *   
 * - ForceMode.Force is analogous to ForceMode.Acceleration but multipled by
 *   mass. It has units of mass*distance/time^2 -- in other words, units of 
 *   force :) Calling AddForce() on each physics frame is equivalent to 
 *   permanently applying the force to the object.
 */

using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class RobotController: MonoBehaviour
{
  private Rigidbody m_rb;

  private void FixedUpdate()
  {
    Vector3 toPlayer = Camera.main.transform.position - transform.position;
    toPlayer.y = 0;
    float distance = toPlayer.magnitude;
    if (distance > 1)// && m_rb.velocity.magnitude < 0.15f)
    {
      m_rb.AddForce(toPlayer.normalized * 0.15f, ForceMode.Acceleration);
    }
  }

  private void Update()
  {
    Vector3 toPlayer = Camera.main.transform.position - transform.position;
    toPlayer.y = 0;
    float distance = toPlayer.magnitude;
    if (distance > 1)
    {
      //m_rb.velocity = toPlayer.normalized * 1f;
    }
  }

  private void Start()
  {
    m_rb = GetComponent<Rigidbody>();
  }
}